"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateJSONSchema = exports.createParser = exports.TransformTarget = void 0;
const definitions_1 = require("./definitions");
var TransformTarget;
(function (TransformTarget) {
    TransformTarget["Encoded"] = "encoded";
    TransformTarget["Decoded"] = "decoded";
})(TransformTarget = exports.TransformTarget || (exports.TransformTarget = {}));
const createParser = (tag, parser) => {
    return {
        tag,
        parse: parser
    };
};
exports.createParser = createParser;
const createPrimitiveParser = (type) => {
    return {
        tag: type,
        parse: () => {
            return { type: type };
        }
    };
};
const StringParser = createPrimitiveParser(definitions_1.CodecType.String);
const NumberParser = createPrimitiveParser(definitions_1.CodecType.Number);
const BooleanParser = createPrimitiveParser(definitions_1.CodecType.Boolean);
const NullParser = createPrimitiveParser(definitions_1.CodecType.Null);
const AnyParser = (0, exports.createParser)(definitions_1.CodecType.Any, (codec) => {
    return {};
});
const EnumParser = (0, exports.createParser)(definitions_1.CodecType.Enum, (codec) => {
    return {
        type: 'string',
        enum: Object.values(codec.props.enum)
    };
});
const LiteralParser = (0, exports.createParser)(definitions_1.CodecType.Literal, (codec) => {
    return {
        type: 'string',
        const: codec.props.value
    };
});
const ObjectParser = (0, exports.createParser)(definitions_1.CodecType.Object, (codec, options) => {
    const entries = Object.entries(codec.props.shape);
    return {
        type: 'object',
        properties: entries.reduce((acc, [key, codec]) => {
            acc[key] = RootParser(codec, options);
            return acc;
        }, {}),
        additionalProperties: !!options?.allowAdditional,
        required: entries
            .filter(([, codec]) => {
            return codec.props.required;
        })
            .map(([key]) => key)
    };
});
const RecordParser = (0, exports.createParser)(definitions_1.CodecType.Record, (codec, options) => {
    return {
        type: 'object',
        additionalProperties: RootParser(codec.props.type, options),
        properties: {},
        required: []
    };
});
const ArrayParser = (0, exports.createParser)(definitions_1.CodecType.Array, (codec, options) => {
    return {
        type: 'array',
        items: RootParser(codec.props.type, options)
    };
});
const TupleParser = (0, exports.createParser)(definitions_1.CodecType.Tuple, (codec, options) => {
    return {
        type: 'array',
        items: codec.props.codecs.map((codec) => RootParser(codec, options)),
        minItems: codec.props.codecs.length,
        maxItems: codec.props.codecs.length
    };
});
/**
 * Intersections are a bit complicated due to the way that additionalProperties works in JSON-Schema. To
 * support this we need to do different operations based on the type of codecs contained by the
 * intersection.
 *
 * If the intersection contains only object schemas then they are all merged under a single schema so
 * that `additionalProperties` applies against them simultaneously.
 *
 * If the intersection contains a mix of object schemas and unions then we need to construct the cartesian
 * product of all unions against all object schemas, combining them under a single `anyOf` definition. Each
 * product will be a merged schema with `additionalProperties` applying against all properties contained
 * within.
 *
 * Note: You can only really intersect object schemas and so we do not check for any other type of schema
 */
const IntersectionParser = (0, exports.createParser)(definitions_1.CodecType.Intersection, (codec, options) => {
    const schemas = codec.props.codecs.map((codec) => RootParser(codec, options));
    const unions = schemas.filter((schema) => !!schema.anyOf);
    const object_schemas = schemas.filter((schema) => schema.type === 'object');
    const mergeObjectSchemas = (a, b) => {
        return {
            type: 'object',
            properties: {
                ...a.properties,
                ...b.properties
            },
            additionalProperties: !!options?.allowAdditional,
            required: [...a.required, ...b.required]
        };
    };
    if (unions.length > 0) {
        return {
            anyOf: unions.reduce((schemas, union) => {
                return union.anyOf.reduce((schemas, us) => {
                    return object_schemas.reduce((schemas, os) => {
                        return schemas.concat(mergeObjectSchemas(os, us));
                    }, schemas);
                }, schemas);
            }, [])
        };
    }
    return {
        type: 'object',
        properties: schemas.reduce((properties, schema) => {
            return {
                ...properties,
                ...(schema.properties || {})
            };
        }, {}),
        additionalProperties: !!options?.allowAdditional,
        required: schemas.reduce((required, schema) => {
            return required.concat(schema.required || []);
        }, [])
    };
});
const UnionParser = (0, exports.createParser)(definitions_1.CodecType.Union, (codec, options) => {
    return {
        anyOf: codec.props.codecs.map((codec) => RootParser(codec, options))
    };
});
const RecursiveParser = (0, exports.createParser)(definitions_1.CodecType.Recursive, (codec, options) => {
    const ref = {
        $ref: `#/definitions/${codec.props.id}`
    };
    const cached = options.cache.get(codec.props.id);
    if (cached) {
        return ref;
    }
    options.cache.set(codec.props.id, {});
    /**
     * The schema is applied to the definition _after_ pushing the definition to the cache
     * to ensure that there are no cache misses from calling RootParser
     */
    options.cache.set(codec.props.id, RootParser(codec.props.resolver(), options));
    return ref;
});
const RootParser = (codec, options) => {
    const parser = options.parsers.find((parser) => parser.tag === codec._tag);
    if (!parser) {
        throw new Error(`No parser configured for codec ${codec._tag}`);
    }
    const schema = parser.parse(codec, options);
    if (codec.props.metadata?.description) {
        return {
            description: codec.props.metadata.description,
            ...schema
        };
    }
    return schema;
};
const generateJSONSchema = (codec, options) => {
    const parsers = [
        ...(options?.parsers || []),
        AnyParser,
        StringParser,
        NumberParser,
        BooleanParser,
        NullParser,
        LiteralParser,
        EnumParser,
        ObjectParser,
        RecordParser,
        ArrayParser,
        TupleParser,
        IntersectionParser,
        UnionParser,
        RecursiveParser
    ];
    const recursion_cache = new Map();
    const schema = RootParser(codec, {
        parsers,
        target: options?.target ?? TransformTarget.Encoded,
        allowAdditional: options?.allowAdditional ?? false,
        cache: recursion_cache
    });
    return {
        definitions: Object.fromEntries(recursion_cache.entries()),
        ...schema
    };
};
exports.generateJSONSchema = generateJSONSchema;
//# sourceMappingURL=generation.js.map